"""
The :mod:`tsfresh.feature_extraction` module contains methods to extract the features from the time series
"""

from tsfresh.feature_extraction.extraction import extract_features
from tsfresh.feature_extraction.settings import FeatureExtractionSettings
