"""
Module with exemplary data sets to play around with.

See for eample the :ref:`quick-start-label` section on how to use them.
"""
from __future__ import absolute_import
from .robot_execution_failures import load_robot_execution_failures, download_robot_execution_failures